/**
 * Mini-Projekt: Flackerne LED an Port B0
 */
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>

#include "parith-15.h"
#include "taster.h"

// Port fr PWM
#define PORT_PWM PORTB  // OC0A   
#define DDR_PWM  DDRB   // OC0A
#define BIT_PWM  0      // OC0A

// Port fr Taster
#define PORT_TAST PORTB  // OC0A   
#define PIN_TAST  PINB   // OC0A   
#define DDR_TAST  DDRB   // OC0A
#define BIT_TAST  4      // OC0A

#define IRQS_PER_SECOND  1000
#define PRESCALE 32

// Wird jede Millisekunde um 1 verringert und bleibt bei 0 stehen
volatile uint8_t count_ms;

enum
{
    MODE_ON,
    MODE_HARD,
    MODE_OFF,
    MODE_SOFT,
    
    MODE_LAST
};

static uint16_t seed __attribute__ ((section (".noinit")));
void __init_seed(void) __attribute__ ((naked, section (".init8")));

static void ioinit(void); 
static void timers_init(void);

/*****************************************************************/
// Timer0: PWM  @ F_CPU/256, d.h. 31.25 kHz
// Timer1: IRQs @ 1kHz, d.h. alle 1ms
void timers_init (void)
{
    // Timer1 in CTC-Mode und prescaled 32 (TCCR1.CS=6)
    TCCR1 = (1 << CTC1) | (1 << CS12) | (1 << CS11) | (0 << CS10); 
    
    // PoutputCompare fr gewnschte Timer1-Frequenz
    OCR1C = (uint8_t) ((uint32_t) F_CPU / IRQS_PER_SECOND / PRESCALE -1);
    OCR1B = 0;
    
    // OutputCompare-Interrupt B fr Timer1
    TIMSK = (1 << OCIE1B);
    
    // PWM @ F_CPU/PWM_RESOLUTION an OC0A (PORT_PWM)
    TCCR0A = (1 << WGM00) | (1 << COM0A1);
    TCCR0B = (0 << CS01) | (1 << CS00); // Prescale = 1
}

/*****************************************************************/
// ISR wird alle 1 ms ausgefhrt
// Verringert `count_ms' um 1 falls noch nicht 0
SIGNAL (SIG_OUTPUT_COMPARE1B)
{
    static taste_t tast = {.key=1};
    uint8_t val = count_ms;
        
    if (val)
        count_ms = val-1;
        
    get_taster (&tast, PIN_TAST & (1 << BIT_TAST));    

    wdt_reset();
}    

/*****************************************************************/
// Initialize hardware
void ioinit (void) 
{
    // high fuse = 0xcc (BOD = 4.3V, WDT=on)
    wdt_enable (WDTO_15MS);
    
    // Alle (nicht verwendeten) Ports sind IN + PullUp
    DDRB  = 0;
    PORTB = 0x3f;
   
    // Enable Timer1 OutputCompare-IRQ
    // Set    Timer0 PWM
    timers_init();

    // OCRA = OUT + LOW
    DDR_PWM  |=  (1 << BIT_PWM);    
    PORT_PWM &= ~(1 << BIT_PWM);    
    
    // Taster = IN + PullUp
    DDR_TAST  &= ~(1 << BIT_TAST);    
    PORT_TAST |=  (1 << BIT_TAST);    
}

/*****************************************************************/
int main (void)
{
    // Initialisiere Hardware
    ioinit (); 

    // Seed fr prandom() setzen
    // seed wurde vor main() durch __init_seed() bestimmt
    set_seed (seed);
    
    // IRQs sind willkommen
    sei();    
    
    uint8_t mode = MODE_HARD;
    
    uint8_t ocr = 0;
    uint8_t o   = 0;
    // main loop
    while (1)
    {
        (void) ocr;
        (void) o;
        
        if (taster)
        {
            taster = 0;
            if (++mode == MODE_LAST)
                mode = 0;
        }
        
        switch (mode)
        {
            case MODE_ON:
                OCR0A = 0xff;
                break;
                
            case MODE_OFF:
                OCR0A = 0;
                break;
                
            case MODE_SOFT:
                // PWM wandert jede ms ein Schritt
                // zum Sollwert ocr
                if (!count_ms)
                {
                    if (o < ocr)    o++;
                    if (o > ocr)    o--;
                    OCR0A = o;
                    // Zhler neu aufziehen
                    count_ms = 1;
                }
                
                // Sollwert erreicht:
                // neuer, zuflliger Wert fr ocr
                if (o == ocr)
                    ocr = prandom();
                break;
            
            case MODE_HARD:

                // Zhler abgelaufen:
                if (!count_ms)
                {
                    // Zhler neu auf 100ms aufziehen
                    count_ms = 100;
                    // Neuer, zuflliger PWM-Wert
                    do
                        ocr = prandom();
                    while (ocr < 20);

                    OCR0A = ocr;
                }
        } // switch (mode)
    } // while (1)
} // main

// Bestimme seed aus zuflligem RAM-Inhalt
// !!! FUNKTION NICHT AUFRUFEN       !!!
// !!! FUNKTION WIRD AUTOMATISCH     !!!
// !!! IN .init8 vor main AUSGEFHRT !!!
void __init_seed (void)
{
	uint16_t s = 0;
	uint16_t *p = (uint16_t*) (RAMEND+1);
	extern uint16_t __noinit_start;

	while (p >= &__noinit_start + 1)
		s ^= * (--p);
        
	seed = s;
}

